unit MainTestApi;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, Buttons,
  db_h;

type
  { NOTE:
    Pure lexical ordering using numbers (both integer and floating point data types) do not work as expected due to the little endian representation
    in x86 systems (hex number $12345678 is stored in memory as the list of bytes $78, $56, $34, $12).
    For example,
      the 32 bits integer   "1" ($00000001) is stored as $01, $00, $00, $00
      the 32 bits integer   "2" ($00000002) is stored as $02, $00, $00, $00
      the 32 bits integer "256" ($00000100) is stored as $00, $01, $00, $00
      the 32 bits integer "257" ($00000101) is stored as $01, $01, $00, $00
    so using it as keys results in an ordering of
      $00, $01, $00, $00 => "256"
      $01, $00, $00, $00 =>   "1"
      $01, $01, $00, $00 => "257"
      $02, $00, $00, $00 =>   "2"
    To prevent this, bytes must be swaped before storing it in the db and after retrieving it.
  }

  PKey = ^TKey;
  TKey = record
    Id: integer;
  end;

  PData = ^TData;
  TData = record
    Name: array[0..50] of AnsiChar;
  end;

  TFormMain = class(TForm)
    Memo: TMemo;
    ChkUseEnv: TCheckBox;
    BtnOpen: TButton;
    BtnClose: TButton;
    BtnStat: TButton;
    BtnLoadData: TButton;
    BtnDeleteAll: TButton;
    BtnListAllByKey: TButton;
    BtnListAll: TButton;
    BtnDestroy: TButton;
    BtnFirst: TButton;
    BtnNext: TButton;
    BtnPrior: TButton;
    BtnLast: TButton;
    BtnDelete: TButton;
    BtnDeleteRandom: TButton;
    BtnCloseCursor: TButton;
    procedure BtnOpenClick(Sender: TObject);
    procedure BtnCloseClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure BtnLoadDataClick(Sender: TObject);
    procedure BtnListAllByKeyClick(Sender: TObject);
    procedure BtnListAllClick(Sender: TObject);
    procedure BtnStatClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure BtnDeleteAllClick(Sender: TObject);
    procedure BtnDestroyClick(Sender: TObject);
    procedure BtnFirstClick(Sender: TObject);
    procedure BtnNextClick(Sender: TObject);
    procedure BtnPriorClick(Sender: TObject);
    procedure BtnLastClick(Sender: TObject);
    procedure BtnDeleteClick(Sender: TObject);
    procedure BtnDeleteRandomClick(Sender: TObject);
    procedure BtnCloseCursorClick(Sender: TObject);
  private
    { Private declarations }
    FEnv: PDB_ENV;
    FDB: PDB;
    FCursor: PDBC;
    procedure UpdateButtons;
    procedure PrintPackedData(const key: PKey; const data: PData); overload;
    procedure PrintPackedData(const key: TKey; const data: TData); overload;
  public
    { Public declarations }
  end;

var
  FormMain: TFormMain;


implementation

{$R *.dfm}

uses
  StrUtils;


{$IFNDEF UNICODE}
type
  { UnicodeString is really a function in Delphi 2009+, but defining it as a type allow the compiler to "eliminate" the call
    in prior (non-Unicode) Delphi versions.
  }
  UnicodeString = string;
{$ENDIF}

function Swap32(i: integer): integer; register;
asm
  bswap eax;
end;

procedure PackKey(var key: TKey);
begin
  key.Id := Swap32(key.id);
end;

procedure UnpackKey(var key: TKey);
begin
  key.Id := Swap32(key.id);
end;

procedure PackData(var key: TKey; var data: TData; id: integer; name: string);
begin
  key.Id := Swap32(id);
  StrLCopy(data.Name, PAnsiChar(AnsiString(name)), SizeOf(data.Name)-1);
end;

procedure ErrCall(dbenv: PDB_ENV; errpfx, msg: PAnsiChar); cdecl;
begin
  MessageDlg(errpfx+': '+string(AnsiString(msg)), mtError, [mbOk], 0);
end;


procedure TFormMain.FormCreate(Sender: TObject);
begin
  InitBerkeleyDB;
  UpdateButtons;
end;

procedure TFormMain.FormDestroy(Sender: TObject);
begin
  BtnCloseClick(nil);
end;

procedure TFormMain.UpdateButtons;
begin
  BtnOpen        .Enabled := (FEnv=nil) and (FDB=nil);
  BtnClose       .Enabled := not BtnOpen.Enabled;
  BtnDestroy     .Enabled := BtnOpen.Enabled;
  ChkUseEnv      .Enabled := BtnOpen.Enabled;
  BtnStat        .Enabled := FDB<>nil;
  BtnLoadData    .Enabled := FDB<>nil;
  BtnDeleteAll   .Enabled := FDB<>nil;
  BtnListAllByKey.Enabled := FDB<>nil;
  BtnListAll     .Enabled := FDB<>nil;
  BtnDeleteRandom.Enabled := FDB<>nil;
  BtnFirst       .Enabled := FDB<>nil;
  BtnNext        .Enabled := FCursor<>nil;
  BtnPrior       .Enabled := FCursor<>nil;
  BtnLast        .Enabled := FDB<>nil;
  BtnDelete      .Enabled := FCursor<>nil;
  BtnCloseCursor .Enabled := FCursor<>nil;
end;

procedure TFormMain.BtnOpenClick(Sender: TObject);
begin
  Memo.Lines.Clear;
  Memo.Lines.Add('Opening...');

  if ChkUseEnv.Checked then begin
    CheckBDB(db_env_create(FEnv, 0));
    try
      FEnv.set_errpfx(FEnv, 'DBENV');
      FEnv.set_errcall(FEnv,ErrCall);

      CheckBDB(FEnv.set_alloc(FEnv, @GetMemory, @ReallocMemory, @FreeMemory));

      CheckBDB(FEnv.open(FEnv, '.', DB_CREATE_ or DB_INIT_MPOOL, 0));
    except
      CallBDBandNil(FEnv.close(FEnv, 0), FEnv);
      raise;
    end;
  end;

  CheckBDB(db_create(FDB, FEnv, 0));
  try
    FDB.set_errpfx(FDB, 'DB');
    FDB.set_errcall(FDB, ErrCall);

    if FEnv=nil then
      CheckBDB(FDB.set_alloc(FDB, @GetMemory, @ReallocMemory, @FreeMemory));

    CheckBDB(FDB.open(FDB, nil, 'Test.db', 'Test', DB_BTREE, DB_CREATE_, 0));
  except
    CallBDBandNil(FDB.close(FDB, 0), FDB);
    raise;
  end;

  Memo.Lines.Add('Opened');
  UpdateButtons;
end;

procedure TFormMain.BtnCloseClick(Sender: TObject);
begin
  if not (csDestroying in ComponentState) then begin
    Memo.Lines.Clear;
    Memo.Lines.Add('Closing...');
  end;

  BtnCloseCursorClick(nil);

  if FDB<>nil then
    CheckBDBandNil(FDB.close(FDB, 0), FDB);

  if FEnv<>nil then
    CheckBDBandNil(FEnv.close(FEnv, 0), FEnv);

  if not (csDestroying in ComponentState) then begin
    Memo.Lines.Add('Closed');
    UpdateButtons;
  end;
end;

procedure TFormMain.BtnDestroyClick(Sender: TObject);
begin
  Memo.Lines.Clear;
  Memo.Lines.Add('Destroying...');

  if ChkUseEnv.Checked then begin
    // destroy database
    CheckBDB(db_env_create(FEnv, 0));
    try
      FEnv.set_errpfx(FEnv, 'DBENV');
      FEnv.set_errcall(FEnv,ErrCall);
      CheckBDB(FEnv.open(FEnv, '.', DB_CREATE_ or DB_INIT_MPOOL, 0));
      CheckBDB(FEnv.dbremove(FEnv, nil, 'Test.db', nil, 0));
    except
      CallBDBandNil(FEnv.close(FEnv, 0), FEnv);
      raise;
    end;
    CheckBDBandNil(FEnv.close(FEnv, 0), FEnv);

    // destroy environment
    CheckBDB(db_env_create(FEnv, 0));
    CheckBDBandNil(FEnv.remove(FEnv, '.', 0), FEnv);
  end
  else begin
    // destroy database
    CheckBDB(db_create(FDB, nil, 0));
    FDB.set_errpfx(FDB, 'DB');
    FDB.set_errcall(FDB, ErrCall);
    CheckBDBandNil(FDB.remove(FDB, 'Test.db', nil, 0), FDB);
  end;

  Memo.Lines.Add('Destroyed');
  UpdateButtons;
end;

procedure TFormMain.BtnStatClick(Sender: TObject);
var
  stat: PDB_BTREE_STAT;
begin
  Memo.Lines.Clear;
  Memo.Lines.Add('Stats...');

  CheckBDB(FDB.Stat(FDB, nil, @stat, 0{DB_FAST_STAT}));
  Memo.Lines.Add('magic        '+IntToStr(stat.bt_magic      ));
  Memo.Lines.Add('version      '+IntToStr(stat.bt_version    ));
  Memo.Lines.Add('metaflags    '+IntToStr(stat.bt_metaflags  ));
  Memo.Lines.Add('nkeys        '+IntToStr(stat.bt_nkeys      ));
  Memo.Lines.Add('ndata        '+IntToStr(stat.bt_ndata      ));
  Memo.Lines.Add('pagesize     '+IntToStr(stat.bt_pagesize   ));
  Memo.Lines.Add('minkey       '+IntToStr(stat.bt_minkey     ));
  Memo.Lines.Add('re_len       '+IntToStr(stat.bt_re_len     ));
  Memo.Lines.Add('re_pad       '+IntToStr(stat.bt_re_pad     ));
  Memo.Lines.Add('levels       '+IntToStr(stat.bt_levels     ));
  Memo.Lines.Add('int_pg       '+IntToStr(stat.bt_int_pg     ));
  Memo.Lines.Add('leaf_pg      '+IntToStr(stat.bt_leaf_pg    ));
  Memo.Lines.Add('dup_pg       '+IntToStr(stat.bt_dup_pg     ));
  Memo.Lines.Add('over_pg      '+IntToStr(stat.bt_over_pg    ));
  Memo.Lines.Add('empty_pg     '+IntToStr(stat.bt_empty_pg   ));
  Memo.Lines.Add('free         '+IntToStr(stat.bt_free       ));
  Memo.Lines.Add('int_pgfree   '+IntToStr(stat.bt_int_pgfree ));
  Memo.Lines.Add('leaf_pgfree  '+IntToStr(stat.bt_leaf_pgfree));
  Memo.Lines.Add('dup_pgfree   '+IntToStr(stat.bt_dup_pgfree ));
  Memo.Lines.Add('over_pgfree  '+IntToStr(stat.bt_over_pgfree));
  FreeMemory(stat);
end;

procedure TFormMain.PrintPackedData(const key: PKey; const data: PData);
var
  unpackedKey: TKey;
  str: string;
begin
  unpackedKey := key^;
  UnpackKey(unpackedKey);
  str := IntToStr(unpackedKey.Id);

  if data<>nil then
    str := str+' / '+UnicodeString(data.Name);

  Memo.Lines.Add(str);
end;

procedure TFormMain.PrintPackedData(const key: TKey; const data: TData);
begin
  PrintPackedData(@key, @data);
end;

procedure TFormMain.BtnLoadDataClick(Sender: TObject);
var
  i: integer;
  key: TKey;
  data: TData;
  dbtKey, dbtData: DBT;
  cnt: integer;
begin
  Memo.Lines.Clear;
  Memo.Lines.Add('Loading data...');

  cnt := 0;

  for i := 1 to 365 do begin
    PackData(key, data, i, FormatDateTime('dddddd', EncodeDate(2011, 1, 1)+i-1));

    FillChar(dbtKey, Sizeof(DBT), 0);
    FillChar(dbtData, Sizeof(DBT), 0);

    dbtKey.data := @key;
    dbtKey.size := SizeOf(key);

    dbtData.data := @data;
    dbtData.size := SizeOf(data);

    if CheckAndNotExistsBDB(FDB.put(FDB, nil, @dbtKey, @dbtData, DB_NOOVERWRITE)) then begin
      PrintPackedData(key, data);
      Inc(cnt);
    end else begin
      UnpackKey(key);
      Memo.Lines.Add(IntToStr(key.Id)+' / ** Already exists **');
    end;
  end;

  Memo.Lines.Add(Format('%d record(s) loaded', [cnt]));
end;

procedure TFormMain.BtnDeleteAllClick(Sender: TObject);
var
  cnt: integer;
begin
  Memo.Lines.Clear;
  Memo.Lines.Add('Deleting...');

  CheckBDB(FDB.truncate(FDB, nil, @cnt, 0));

  Memo.Lines.Add(Format('%d record(s) deleted', [cnt]));
end;

procedure TFormMain.BtnListAllClick(Sender: TObject);
var
  cursor: PDBC;
  dbtKey, dbtData: DBT;
  cnt: integer;
begin
  Memo.Lines.Clear;
  Memo.Lines.Add('Listing...');

  CheckBDB(FDB.cursor(FDB, nil, @cursor, 0));
  try
    FillChar(dbtKey, Sizeof(DBT), 0);
    FillChar(dbtData, Sizeof(DBT), 0);

    cnt := 0;

    while CheckAndFoundBDB(cursor.get(cursor, @dbtKey, @dbtData, DB_NEXT)) do begin
      PrintPackedData(PKey(dbtKey.data)^, PData(dbtData.data)^);
      Inc(cnt);
    end;
  finally
    CheckBDB(cursor.close(cursor));
  end;

  Memo.Lines.Add(Format('%d record(s) listed', [cnt]));
end;

procedure TFormMain.BtnListAllByKeyClick(Sender: TObject);
var
  i: integer;
  key: TKey;
  dbtKey, dbtData: DBT;
  cnt: integer;
begin
  Memo.Lines.Clear;
  Memo.Lines.Add('List by key...');

  FillChar(dbtKey, Sizeof(DBT), 0);
  FillChar(dbtData, Sizeof(DBT), 0);

  cnt := 0;

  for i := 1 to 365 do begin
    key.Id := i;
    PackKey(key);

    dbtKey.data := @key;
    dbtKey.size := SizeOf(key);

    if CheckAndFoundBDB(FDB.get(FDB, nil, @dbtKey, @dbtData, 0)) then begin
      PrintPackedData(key, PData(dbtData.data)^);
      Inc(cnt);
    end else begin
      UnpackKey(key);
      Memo.Lines.Add(IntToStr(key.Id)+' / ** Not found **');
    end;
  end;

{ ALTERNATIVE:

  FillChar(dbtKey, Sizeof(DBT), 0);
  FillChar(dbtData, Sizeof(DBT), 0);

  for i := 1 to 365 do begin
    key.Id := i;
    PackKey(key);

    dbtKey.data := @key;
    dbtKey.size := SizeOf(key);

    dbtData.data := @data;
    dbtData.ulen := SizeOf(data); // size of user (provided) memory
    dbtData.flags := DB_DBT_USERMEM;

    CheckBDB(FDB.get(FDB, nil, @dbtKey, @dbtData, 0));

    PrintPackedData(key, data);
  end;
}

  Memo.Lines.Add(Format('%d record(s) listed', [cnt]));
end;

procedure TFormMain.BtnDeleteRandomClick(Sender: TObject);
var
  key: TKey;
  dbtKey: DBT;
begin
  Memo.Lines.Clear;
  Memo.Lines.Add('Deleting...');

  FillChar(dbtKey, Sizeof(DBT), 0);
  Randomize;
  key.Id := Random(365)+1;
  PackKey(key);
  dbtKey.data := @key;
  dbtKey.size := SizeOf(key);
  PrintPackedData(dbtKey.data, nil);
  CheckBDB(FDB.del(FDB, nil, @dbtKey, 0));

  Memo.Lines.Add('Deleted');
  UpdateButtons;
end;

procedure TFormMain.BtnFirstClick(Sender: TObject);
var
  dbtKey, dbtData: DBT;
begin
  Memo.Lines.Clear;

  if FCursor=nil then
    CheckBDB(FDB.cursor(FDB, nil, @FCursor, 0));

  FillChar(dbtKey, Sizeof(DBT), 0);
  FillChar(dbtData, Sizeof(DBT), 0);

  CheckBDB(FCursor.get(FCursor, @dbtKey, @dbtData, DB_FIRST));
  PrintPackedData(PKey(dbtKey.data)^, PData(dbtData.data)^);

  UpdateButtons;
end;

procedure TFormMain.BtnNextClick(Sender: TObject);
var
  dbtKey, dbtData: DBT;
begin
  Memo.Lines.Clear;

  FillChar(dbtKey, Sizeof(DBT), 0);
  FillChar(dbtData, Sizeof(DBT), 0);

  CheckBDB(FCursor.get(FCursor, @dbtKey, @dbtData, DB_NEXT));
  PrintPackedData(PKey(dbtKey.data)^, PData(dbtData.data)^);

  UpdateButtons;
end;

procedure TFormMain.BtnPriorClick(Sender: TObject);
var
  dbtKey, dbtData: DBT;
begin
  Memo.Lines.Clear;

  FillChar(dbtKey, Sizeof(DBT), 0);
  FillChar(dbtData, Sizeof(DBT), 0);

  CheckBDB(FCursor.get(FCursor, @dbtKey, @dbtData, DB_PREV));
  PrintPackedData(PKey(dbtKey.data)^, PData(dbtData.data)^);

  UpdateButtons;
end;

procedure TFormMain.BtnLastClick(Sender: TObject);
var
  dbtKey, dbtData: DBT;
begin
  Memo.Lines.Clear;

  if FCursor=nil then
    CheckBDB(FDB.cursor(FDB, nil, @FCursor, 0));

  FillChar(dbtKey, Sizeof(DBT), 0);
  FillChar(dbtData, Sizeof(DBT), 0);

  CheckBDB(FCursor.get(FCursor, @dbtKey, @dbtData, DB_LAST));
  PrintPackedData(PKey(dbtKey.data)^, PData(dbtData.data)^);

  UpdateButtons;
end;

procedure TFormMain.BtnDeleteClick(Sender: TObject);
var
  dbtKey, dbtData: DBT;
begin
  Memo.Lines.Clear;
  Memo.Lines.Add('Deleting...');

  FillChar(dbtKey, Sizeof(DBT), 0);
  FillChar(dbtData, Sizeof(DBT), 0);
  CheckBDB(FCursor.get(FCursor, @dbtKey, @dbtData, DB_CURRENT));
  PrintPackedData(PKey(dbtKey.data)^, PData(dbtData.data)^);

  CheckBDB(FCursor.del(FCursor, 0));

  Memo.Lines.Add('Deleted');
  UpdateButtons;
end;

procedure TFormMain.BtnCloseCursorClick(Sender: TObject);
begin
  if FCursor<>nil then
    CheckBDBandNil(FCursor.close(FCursor), FCursor);

  UpdateButtons;
end;

end.
